<?php
// Own namespace
namespace Org\Mxchange\CoreFramework\Database\Frontend\User;

// Import framework stuff
use Org\Mxchange\CoreFramework\Bootstrap\FrameworkBootstrap;
use Org\Mxchange\CoreFramework\Database\Frontend\BaseDatabaseFrontend;
use Org\Mxchange\CoreFramework\Factory\Object\ObjectFactory;
use Org\Mxchange\CoreFramework\Frontend\Account\ManageableAccountFrontend;
use Org\Mxchange\CoreFramework\Generic\NullPointerException;
use Org\Mxchange\CoreFramework\Registration\User\UserRegister;
use Org\Mxchange\CoreFramework\Registry\Registerable;
use Org\Mxchange\CoreFramework\Result\Update\UpdateableResult;

/**
 * A database frontend for the User class
 *
 * @author		Roland Haeder <webmaster@shipsimu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2021 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.shipsimu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
class UserDatabaseFrontend extends BaseDatabaseFrontend implements ManageableAccountFrontend, Registerable {
	// Constants for exceptions
	const EXCEPTION_CLIENT_USERNAME_NOT_FOUND = 0x180;

	// Constants for database columns
	const DB_COLUMN_USERID       = 'userid';
	const DB_COLUMN_USERNAME     = 'username';
	const DB_COLUMN_EMAIL        = 'email';
	const DB_COLUMN_CONFIRM_HASH = 'confirm_hash';
	const DB_COLUMN_USER_STATUS  = 'user_status';

	// Constants for database table names
	const DB_TABLE_USER = 'user';

	/**
	 * Protected constructor
	 *
	 * @return	void
	 */
	private function __construct () {
		// Call parent constructor
		parent::__construct(__CLASS__);
	}

	/**
	 * Creates an instance of this database frontend by a provided user class
	 *
	 * @return	$frontendInstance	An instance of the created frontend class
	 */
	public static final function createUserDatabaseFrontend () {
		// Get a new instance
		$frontendInstance = new UserDatabaseFrontend();

		// Set (primary!) table name
		$frontendInstance->setTableName(self::DB_TABLE_USER);

		// Return the instance
		return $frontendInstance;
	}

	/**
	 * Handles inserting the registration data from a registration instance into the database
	 *
	 * @param	$registrationInstance	An instance of a registration class
	 * @return	void
	 */
	public function insertRegistrationObject (UserRegister $registrationInstance) {
		// Generate a data set for the request
		$dataSetInstance = ObjectFactory::createObjectByConfiguredName('dataset_criteria_class', array(self::DB_TABLE_USER));

		// Set the primary key
		$dataSetInstance->setUniqueKey(self::DB_COLUMN_USERNAME);

		// Add registration elements to the dataset
		$registrationInstance->addElementsToDataSet($dataSetInstance);

		// "Insert" this request instance completely into the database
		$this->queryInsertDataSet($dataSetInstance);
	}

	/**
	 * Updates an user database entry with given result
	 *
	 * @param	$resultInstance		An instance of a UpdateableResult class
	 * @return	void
	 * @throws	NullPointerException	If $updateInstance or $searchInstance is null
	 */
	public function doUpdateByResult (UpdateableResult $resultInstance) {
		// Get the search instance from result
		$searchInstance = $resultInstance->getSearchInstance();

		// Is this null?
		if (is_null($searchInstance)) {
			// Get the update instance
			$updateInstance = $resultInstance->getUpdateInstance();

			// Is this null?
			if (is_null($updateInstance)) {
				// Throw an exception here
				throw new NullPointerException($this, self::EXCEPTION_IS_NULL_POINTER);
			}

			// Get search instance from update instance
			$searchInstance = $updateInstance->getSearchInstance();

			// Is it still null?
			if (is_null($searchInstance)) {
				// Throw an exception here
				throw new NullPointerException($updateInstance, self::EXCEPTION_IS_NULL_POINTER);
			}
		}

		// Generate a data set object
		$dataSetInstance = ObjectFactory::createObjectByConfiguredName('dataset_criteria_class', array(self::DB_TABLE_USER));

		// Add seach criteria
		$dataSetInstance->setSearchInstance($searchInstance);

		// Set the primary key
		$dataSetInstance->setUniqueKey(self::DB_COLUMN_USERNAME);

		// Add all update criteria to the database set
		$resultInstance->addElementsToDataSet($dataSetInstance);

		// "Update" this request with the database
		FrameworkBootstrap::getDatabaseInstance()->queryUpdateDataSet($dataSetInstance);
	}

}
