<?php
// Own namespace
namespace Org\Mxchange\CoreFramework\Response;

// Import framework stuff
use Org\Mxchange\CoreFramework\Bootstrap\FrameworkBootstrap;
use Org\Mxchange\CoreFramework\Helper\Application\ApplicationHelper;
use Org\Mxchange\CoreFramework\Object\BaseFrameworkSystem;
use Org\Mxchange\CoreFramework\Registry\GenericRegistry;

/**
 * A generic request class
 *
 * @author		Roland Haeder <webmaster@shipsimu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2021 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.shipsimu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * The extended headers are taken from phpMyAdmin setup tool, written by
 * Michal Cihar <michal@cihar.com>, licensed under GNU GPL 2.0.
 */
abstract class BaseResponse extends BaseFrameworkSystem {
	/**
	 * Response status
	 */
	private $responseStatus = '200 OK';

	/**
	 * Array with all headers
	 */
	private $responseHeaders = [];

	/**
	 * Cookies we shall sent out
	 */
	private $cookies = [];

	/**
	 * Body of the response
	 */
	private $responseBody = '';

	/**
	 * Response type
	 */
	private $responseType = 'invalid';

	/**
	 * Protected constructor
	 *
	 * @param	$className	Name of the concrete response
	 * @return	void
	 */
	protected function __construct (string $className) {
		// Call parent constructor
		parent::__construct($className);
	}

	/**
	 * Setter for status
	 *
	 * @param	$status		New response status
	 * @return	void
	 */
	public final function setResponseStatus (string $status) {
		$this->responseStatus = $status;
	}

	/**
	 * Add header element
	 *
	 * @param	$name	Name of header element
	 * @param	$value	Value of header element
	 * @return	void
	 */
	public final function addHeader (string $name, $value) {
		$this->responseHeaders[$name] = $value;
	}

	/**
	 * Reset the header array
	 *
	 * @return	void
	 */
	public final function resetResponseHeaders () {
		$this->responseHeaders = [];
	}

	/**
	 * "Writes" data to the response body
	 *
	 * @param	$output		Output we shall sent in the HTTP response
	 * @return	void
	 */
	public final function writeToBody (string $output) {
		$this->responseBody .= $output;
	}

	/**
	 * Sets the response body to something new
	 *
	 * @param	$output		Output we shall sent in the HTTP response
	 * @return	void
	 */
	public final function setResponseBody ($output) {
		$this->responseBody = $output;
	}

	/**
	 * Setter for response type
	 *
	 * @param	$responseType	Response type
	 * @return	void
	 */
	protected final function setResponseType (string $responseType) {
		$this->responseType = $responseType;
	}

	/**
	 * Getter for response type
	 *
	 * @param	$responseType	Response type
	 * @return	void
	 */
	public final function getResponseType () {
		return $this->responseType;
	}

	/**
	 * Adds a fatal message id to the response. The added messages can then be
	 * processed and outputed to the world
	 *
	 * @param	$messageId	The message id we shall add
	 * @return	void
	 */
	public final function addFatalMessage (string $messageId) {
		// Adds the resolved message id to the fatal message list
		$this->addFatalMessagePlain(FrameworkBootstrap::getLanguageInstance()->getMessage($messageId));
	}

	/**
	 * Adds a plain fatal message id to the response
	 *
	 * @param	$message	The plain message we shall add
	 * @return	void
	 */
	public final function addFatalMessagePlain (string $message) {
		// Adds the resolved message id to the fatal message list
		$this->pushValueToGenericArrayKey('fatal_messages', 'generic', 'message', $message);
	}

	/**
	 * Flushs the cached HTTP response to the outer world
	 *
	 * @param	$force	Whether we shall force the output or abort if headers are
	 *					already sent with an exception
	 * @return	void
	 * @throws	ResponseHeadersAlreadySentException		Thrown if headers are
	 *													already sent
	 */
	public function flushBuffer (bool $force = false) {
		// Get application instance
		$applicationInstance = ApplicationHelper::getSelfInstance();

		// Headers already sent?
		if ((headers_sent()) && ($force === false)) {
			// Headers are already sent!
			throw new ResponseHeadersAlreadySentException($this, self::EXCEPTION_HEADERS_ALREADY_SENT);
		} elseif (!headers_sent()) {
			// Send headers out
			header('HTTP/1.1 ' . $this->responseStatus);

			// Used later
			$now = gmdate('D, d M Y H:i:s') . ' GMT';

			// General header for no caching
			$this->addHeader('Expired', $now); // RFC2616 - Section 14.21
			$this->addHeader('Last-Modified', $now);
			$this->addHeader('Cache-Control', 'no-store, no-cache, must-revalidate, pre-check=0, post-check=0, max-age=0'); // HTTP/1.1
			$this->addHeader('Pragma', 'no-cache'); // HTTP/1.0

			// Define the charset to be used
			//$this->addHeader('Content-type:', sprintf("text/html; charset=%s", FrameworkBootstrap::getConfigurationInstance()->getConfigEntry('header_charset')));

			// Send all headers
			foreach ($this->responseHeaders as $name => $value) {
				header($name . ': ' . $value);
				//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('name=' . $name . ',value=' . $value);
			}

			// Send cookies out?
			if (count($this->cookies) > 0) {
				// Send all cookies
				$cookieString = implode(' ', $this->cookies);
				header('Set-Cookie: ' . $cookieString);

				// Remove them all
				$this->cookies = [];
			}
		}

		// Are there some error messages?
		if ((!$this->isValidGenericArrayKey('fatal_messages', 'generic', 'message')) || ($this->countGenericArrayElements('fatal_messages', 'generic', 'message') == 0)) {
			// Flush the output to the world
			$this->getWebOutputInstance()->output($this->responseBody);
		} else {
			// Display all error messages
			$applicationInstance->handleFatalMessages($this->getGenericArrayKey('fatal_messages', 'generic', 'message'));

			// Send the error messages out to the world
			$this->getWebOutputInstance()->output($this->responseBody);
		}

		// Clear response header and body
		$this->setResponseBody('');
		$this->resetResponseHeaders();
	}

	/**
	 * "Getter" for default command
	 *
	 * @return	$defaultCommand		Default command for this response
	 */
	public function determineDefaultCommand () {
		// Get application instance
		$applicationInstance = ApplicationHelper::getSelfInstance();

		// Generate config key
		$configKey = sprintf('default_%s_%s_command',
			$applicationInstance->getAppShortName(),
			$this->getResponseType()
		);

		// Get default command response
		$defaultCommand = FrameworkBootstrap::getConfigurationInstance()->getConfigEntry($configKey);

		// Return it
		return $defaultCommand;
	}

}
